# -*- coding: utf-8 -*-

# Copyright 2024 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .builder import Builder
import openmdao.api as om
import numpy as np
import scipy.interpolate as spip

class RbfXferSolver(om.ExplicitComponent):
    """OpenMDAO interpolator wrapper
    """
    def initialize(self):
        self.options.declare('sol', desc='NIPK solver', recordable=False)

    def setup(self):
        self.sol = self.options['sol']
        self.add_input('x_struct0', shape_by_conn=True, desc='structural coordinates')
        self.add_input('q_struct', shape_by_conn=True, desc='modal displacements on structural grid')
        self.add_input('x_aero0', shape_by_conn=True, desc='aerodynamic coordinates')
        self.add_output('q_aero', shape=(3 * self.sol.n_nodes_aero, self.sol.n_modes), desc='modal displacements on aerodynamic grid')

    def compute(self, inputs, outputs):
        outputs['q_aero'] = self.sol.compute(inputs['x_struct0'], inputs['q_struct'], inputs['x_aero0'])

class RbfXferBuilder(Builder):
    """Rbf interpolator
    """
    def __init__(self, struct_builder, aero_builder, cfg):
        n_s = struct_builder.get_number_of_nodes() # number of structural nodes
        n_a = aero_builder.get_number_of_nodes() # number of aerodynamic nodes
        self.xfer = RbfInterp(cfg, n_s, n_a)

    def get_solver(self, scenario_name=''):
        """Return OpenMDAO component containing the solver
        """
        return RbfXferSolver(sol=self.xfer)

class RbfInterp():
    """Scipy RBF interpolator
    """
    def __init__(self, cfg, n_s, n_a):
        self.n_nodes_struct = n_s # number of structural nodes
        self.n_nodes_aero = n_a # number of aerodynamic nodes
        self.n_dim = cfg['n_dim'] # number of coordinates to use for interpolation
        self.n_modes = cfg['n_modes'] # number of structural modes
        self.smoothing = cfg['smooth'] # smoothing parameter

    def compute(self, x_s, q_s, x_a):
        # Unflatten inputs
        x_struct = np.zeros((self.n_nodes_struct, 3))
        q_struct = np.zeros((self.n_nodes_struct, 3 * self.n_modes))
        x_aero = np.zeros((self.n_nodes_aero, 3))
        for i in range(self.n_nodes_struct):
            for j in range(3):
                x_struct[i, j] = x_s[3 * i + j]
        for i in range(self.n_nodes_struct):
            for j in range(self.n_modes):
                for k in range(3):
                    q_struct[i, 3 * j + k] = q_s[3 * i + k, j]
        for i in range(self.n_nodes_aero):
            for j in range(3):
                x_aero[i, j] = x_a[3 * i + j]
        # Compute
        q_aero = spip.RBFInterpolator(x_struct[:,:self.n_dim], q_struct, smoothing=self.smoothing)(x_aero[:,:self.n_dim])
        # Flatten output
        q_a = np.zeros((3 * self.n_nodes_aero, self.n_modes))
        for i in range(self.n_nodes_aero):
            for j in range(self.n_modes):
                for k in range(3):
                    q_a[3 * i + k, j] = q_aero[i, 3 * j + k]
        return q_a
